block_time = 60 #in seconds
max_target_increase = 1.2
max_target_decrease = 0.8
initial_target = 2**240

COIN = 100000000
initial_reward = 600*COIN
reward_decrease_halflife = 2100000 # in blocks

output_creation_fee = 22000

coinbase_maturity = 60

serialized_genesis_block = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x08\xcd\xda\xcf\xa0X\x81\xdd\xfb\n\x9e\x10M\xf3!\xd6\xb5\xb1\x1eU\xa6\x98\xc54,1\xab\xa4\xaa\xf8SeU\xe1te\x90o\xde\xa3:-\xf8\xfa\xe8=\xe5\xdf\x8e\x84m!^\xa3$D\x89\xe2\x91\xa0k\xee\x01\xaef\xcd5^\xac\x15\x99\x1aJ\xfc\x06f\xbcE\x86\x00\x9b\xd2H\xc3\xb1\xfd\x01\\\xca6[_tPh\xdd\xc7\x03q\x11\xb0\xa4\xc0\xbfVE\xbaK\xb9EQS\x00j\x0e\xad!\x08\x13\x1d\xb7F\xda\x80\x81\xa8~I\n\xce\x99\\-\xdaD\xa2\xcf-\xfd3\x01\x7f\x0f^i\x08\x7f5\xe0\x83\x95\xe7\xaeQ\xe3\xcf\xf2s\xdfg\x96W\x00\x00\x00\r\xf8GX\x00\x00[\xff\x97\x00\x00\xf0\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01z\xa9\x03\x00\x00\x00\x01\x00\x01\x08\xcd\xda\xcf\xa0X\x81\xdd\xfb\n\x9e\x10M\xf3!\xd6\xb5\xb1\x1eU\xa6\x98\xc54,1\xab\xa4\xaa\xf8SeU\x8c\xf2\x9c\xf5\xf5p\xbf\x0b\x95Y\xea\xe9S}\x9el\x18\xb8\x17\xddN\x8eY=\x97d\xb4\x98\xa5\xe5(\x03\x00\x00\x00\x00\x80A\xe9\xf6\xf7\xc0\xfb\x0c\xe6\xdcE\xb3B\xa2%ce\xfd\x99\r\xbf\x9d\x1fA\xf7ND\xcf\x92^\x81\xf6\xf75O\xf1\xe8x\xf7\xe1\x7f\xc6\xc3@\xd5\x946`\xfc\xe2\xf0~E)\xd8Y\x0c\x19\xf9Jd\xe9T\xda\xe6\x00!\x08\xcd\xda\xcf\xa0X\x81\xdd\xfb\n\x9e\x10M\xf3!\xd6\xb5\xb1\x1eU\xa6\x98\xc54,1\xab\xa4\xaa\xf8SeU\x00\x01\x00\x00\x00\x01\x00\x00\x00<\x0bN\x13\x90\xb0bl\xf8j\xc8/q\xc4!\x06\x1c;\x19\xf2\xd3\xd0\xdc\xb1\x97u\xc6&\xb5\xc7]/\xda\xa0\x00\x00\x00\x00\x00\x00\x00\x00\x08\xcd\xda\xcf\xa0X\x81\xdd\xfb\n\x9e\x10M\xf3!\xd6\xb5\xb1\x1eU\xa6\x98\xc54,1\xab\xa4\xaa\xf8SeU\x019\xbe\xd17\xbevE\xa0\x8a\x00\xed=\xbb\xe6\xe6\x88|\xb0\xd2\x95\x80\xa2\xa4\x0e\xf0\xb9\xc5g\xb9\xd1 \x88#2\xaa\xb0\xc0\xa7;\x99\x07u\xdbu\xa0\x8a\x1a,\x9bn\x01\xdbV/\x8a\xbb\xa8]a\xd9\x8d\xee\xaf\xc9\x00Y\x03\xa2\xf9\x88z\xe2s\xd7~\xb97\xe5\xbd\xd3a\xa0\xde/!\xc1\x1f\x1a?\x08\xdb\xf4f\xe3`6B\x17.\xff\xf36\xc3\x1f\xcd\xb8\x13\xf3\xb4\x00\x9e\xefU\xe7g\xa5N\r\x8fl\x1e\x0b\xf5\x98\xa3\x1bE\xb4\xb3\xb0\xac\n\x836\xa7\x94]8\rq\x82\x9f\x93\xbdf\xc4s.\x0b\x7f\xfcz>\xb5\xd0\x00I \x00\x00\x00\r\xf8GX\x00j\xf1\xdb\xeb\x05?\x0bb\xb5\n\x03vk\xcc\xb6\xedekwF\xfak\x8d\x110\xd20\x12\x89\x1b>\xc4\x8d\x15grp?\x83\xc8\xe2k/q\\\x9c\x80O\x8fT;\xe7\x14\xf5\x91Ktg\xa5s\x13_\x8a\r"
